
/* 
EA7ARX BITX HF Software Control 
Copyright (C) 2005  Manuel Sanchez

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation.

This program is distributed in the hope that it will be useful, but WITHOUT 
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with 
this program; if not, write to the Free Software Foundation, Inc., 51 Franklin 
Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

/********************************************
 * Control PLL    		for QRP USE         *
 *  Version 2.1                             *
 *  29 de Agosto de 2005                    *
 *                                          *
 ********************************************
 Description:
 This is software control for PLL syntethizer based in Motorola MC145170, 
 rotative encoder, six buttons, an external D/A converter and LCD display

 Features:

 - PLL Control.
 - Add FI Frequency to show frequency
 - Non volatile storage of frequency and Mode.
 - 16 chanells can be used as VFOs
 - 2 VFOs, Call frequency
 - Multiband, Multimode selection.
 - 11 bars S-METER

******************************************************************************
 Autor - Manuel Snchez Raya - EA7ARX
 from a work of - Curtis W. Preuss - WB2V y John Langner, WB2OSZ

 Modification History
    8/19/98 - Version 1 - Initial Version by Curtis W. Preuss - WB2V
   12/xx/98 - Version 2 - Converted to MPASM by Bruce Stough, AA0ED 
    4/21/99 - Version 3 - Fixed and modified by Bruce Stough, AA0ED and
                            Craig Johnson, AA0ZZ
	8/07/03 - Version 4 - Features update and pin modifications in assembler 
							by Manuel Snchez Raya EB7HPM. 
    4/01/04 - Version 0b - First Try to program PLL in C failed.
   10/08/05 - Version 1.0 - Frequency control works!!
   18/08/05 - Version 2.0 - Droped several features due to space exahusted. :-(
   29/08/05 - Version 2.3 - First production revised version.
*/


/****************************************************************************
 * Device type and options.                                                 *
 ****************************************************************************/

#define XTAL    4000000         /* Xtal frequency.  Need for */
								/* various timings. */
#include <pic.h>
/*------------------------------------------------------------------------
 *
 * Certain options must be selected when programming the PIC.
 * 
 * Select
 *      code protect off,
 *      power up timer on,
 *      watch dog timer off,
 *      MCLR off,
 *      LVP off,
 *      XT oscillator.
 *
 *-----------------------------------------------------------------------*/

#if _16F628
 
__CONFIG(0011111111110001B);

#else
#error Wrong device type.
#endif

// VFO modes list
#define OPER_VFOA 0
#define OPER_VFOB 1
#define OPER_CANAL 2
#define OPER_CALL 3
#define OPER_SPLIT 4
// Several commun definitions
#define TRUE    1
#define FALSE   0
#define byte_t unsigned char
// Mark to test internal EEPROM
#define MARCA 	0xAF
// Address to information in EEPROM
// Note that all EEPROM memory is used
#define AD_VFOA 	0
#define AD_VFOB 	5
#define AD_CALL 	10
#define AD_FI		15
#define AD_CANAL	35
#define AD_OPER		36
#define AD_CURSOR	37
#define AD_CHNL		38
#define AD_MARCA	127
// digits show in LCD
#define DIGITS 7

/*
 * Functions prototipes follow
 */
void adjust_rutine(void);

/* BCD number manipulation. */
void bcd_clear (byte_t *bcd);
void bcd_incr (unsigned char bcd[], byte_t valor);
void add_fi (unsigned char bcd[], unsigned char valor[]);
void bcd_decr (unsigned char bcd[], byte_t valor);
void bcd_display (void);

/* Multi byte binary number arithmetic. */
void long_add (byte_t *psum, byte_t *padd);
void long_negate (byte_t *pnum);

/* Hardware specific. */
void apagar(void);
void delay (byte_t ms);
void enc_sample (void);
void twist_event (void);
void button_event (void);

/* Parameters EEPROM storage */
byte_t data_eeprom_read (byte_t adr);
void data_eeprom_write(unsigned char,unsigned char);
void load_parameters(void);
void init_parameters(void);
void save_parameters(void);
void save_bcd_packet(unsigned char bcd[],unsigned char pos);
void load_bcd_packet(unsigned char bcd[],unsigned char pos);

/* Liquid Crystal Display. */
void lcd_init (void);
void lcd_put4 (byte_t x);
void lcd_write_data (byte_t x);
void lcd_write_instr (byte_t x);
void lcd_set_cursor (byte_t position);
void lcd_show_cursor (void);
void lcd_hide_cursor (void);

/* Phased Loked Loop. */
void pll_init (void);
void pll_from_bcd(byte_t *f);
void pll_write (unsigned int bcd);
void pll_write_freq (unsigned int f);
void pll_shift_byte (byte_t x);
void dac_write(byte_t);
void bpf_write (unsigned int);

/* 4 bits A/D converter. */
void adc_init ();
unsigned char lee_adc(void);


/*****************************************************************************

 Target Controller -      PIC16F628
                          __________
          DATAS/B5 --RA2 |1       18| RA1--------- RS/B4
     	 LCD ENA ----RA3 |2       17| RA0--------- S-METER
         CLK_PLL ----RA4 |3       16| OSC1--------XTAL
   		   PTT ----!MCLR |4       15| OSC2--------XTAL
     Ground----------Vss |5       14| VDD---------+5 V
     CODFA ----------RB0 |6       13| RB7---------LCD/B3
    ENA_DAC ---------RB1 |7       12| RB6---------LCD/B2
    ENA_BPF ---------RB2 |8       11| RB5---------LCD/B1
     CODFB ----------RB3 |9       10| RB4---------LCD/B0
                          ----------
*/

volatile static bit     EN @ (unsigned)&PORTB*8+4;
volatile static bit     EN_DAC  @ (unsigned)&PORTB*8+1;
volatile static bit     EN_BPF  @ (unsigned)&PORTB*8+2;
volatile static bit     CLK  @ (unsigned)&PORTA*8+4;
volatile static bit     DIN  @ (unsigned)&PORTA*8+2;

volatile static bit		PB0 @ (unsigned)&PORTB*8+4;
volatile static bit     PB1 @ (unsigned)&PORTB*8+5;
volatile static bit		PB2	@ (unsigned)&PORTB*8+6;
volatile static bit     PB3  @ (unsigned)&PORTB*8+7;
volatile static bit     PB4  @ (unsigned)&PORTA*8+1;
volatile static bit     PB5  @ (unsigned)&PORTA*8+2;

volatile static bit     TX_RX  @ (unsigned)&PORTA*8+5;
volatile static bit     LCD_rs  @ (unsigned)&PORTA*8+1;
volatile static bit     LCD_e   @ (unsigned)&PORTA*8+3;

// Ports input/output redefining
#define PORT_B_OUT      0x09  // all output most of the time. 
#define PORT_A_OUT      0x01  // TX_RX bit input. 
#define PORT_B_TECLA    0xF9  // all inputs
#define PORT_A_TECLA    0x07  // all inputs

/* 
 Variables follow
*/

// Main VFO
byte_t vfo[7];
signed char canal,banda;
unsigned char apaga,apagado,mod_packet;
byte_t mode;
byte_t oper;
byte_t cursor_digit;
byte_t freqin[7];
volatile static signed char Twisted;
/*
 Constants follow
*/
//underline step position in LCD
static const unsigned char cur_pos[]={12,10,9,8,6,5};
//Bands conmutation list
unsigned char lista_banda[]={0x35,0x70,0x14,0x21,0x28,0x50};
//Modes list
static const unsigned char mode_t[6][3]={{' ','C','W'},{' ','A','M'},{'U','S','B'},{'L','S','B'},{' ','F','M'}};
//VFO list
static const unsigned char oper_t[]={'A','B','N','C','S'};
//Temporal storage
unsigned char temp[7];

void main ()
{
unsigned char tmp,i,ant_tmp,ant_TX_RX,adc_slope;
/*
 * Set up I/O ports.
 */
        PORTB = 0; 
        TRISB = PORT_B_OUT;

/*
 * Set up timer for interrupts of about 2400 Hz.
 */

 	asm("CLRWDT");		/* Manual says so. */

 	OPTION = 0;		/* Prescale of 2. */
 				/* which is really Fosc/8. */
 				/* Enable weak pullups on any */
 				/* port B inputs. */
 	INTCON = 0;
 	T0IE = 1;
 	GIE = 1; 
	RBPU = 1; //Activa weak pull-ups
/*
 * Initialize the PLL & DAC chips.
 */
	CM2=1;
	CM1=1;
	CM0=1;
	VREN=0;
    PORTA = 0;
    TRISA = PORT_A_OUT;
    pll_init ();
/*
 * Initialize the LCD.
 */
        lcd_init ();
#if 0
        lcd_set_cursor (0);
        lcd_write_data ('E');
        lcd_write_data ('A');
        lcd_write_data ('7');
        lcd_write_data ('A');
        lcd_write_data ('R');
        lcd_write_data ('X');
        lcd_write_data (' ');
        lcd_write_data ('B');
        lcd_write_data ('I');
        lcd_write_data ('T');
        lcd_write_data ('X');
        lcd_write_data (' ');
        lcd_write_data ('M');
        lcd_write_data ('U');
        lcd_write_data ('L');
        lcd_write_data ('T');
        delay(250);
        delay(250);
        delay(250);
        delay(250);
#endif

/*
 * Load Stuff from EEPROM if loaded.
 */
// search for mark, if not initialize all parameters
	if (EEPROM_READ(AD_MARCA)!=MARCA) init_parameters();
	else load_parameters();

/*
 * Wait for user input.
 */
 	Twisted = 0; 
	ant_tmp=0;
	adc_slope=0;
	ant_TX_RX=1;
	mod_packet=0;
	lcd_set_cursor (16);
 	lcd_write_data ('S');
	apaga=1;
	bcd_display();
	PORTB=0xF9;
    TRISB = PORT_B_TECLA; 	
    TRISA = PORT_A_TECLA; 

// Power up protection
	if (!PB3) apagado=0;
	else apagado=1;
/*
 * Next: Was adjust or reset invoked?
 */
	if (!PB4) adjust_rutine();
	//mod_packet manages two storage formats in EEPROM
	mod_packet=1;
	if (!PB5) init_parameters();

    for (;;) {
/*
 * First priority: change frequency if SPLIT mode.
 */
	if (TX_RX!=ant_TX_RX) {
	    TRISB = PORT_B_OUT;
	    TRISA = PORT_A_OUT;
		ant_TX_RX=TX_RX;
		if (oper==OPER_SPLIT) {
			if (TX_RX) load_bcd_packet(vfo,0);
			else load_bcd_packet(vfo,4);
		}
	  	bcd_display ();	
		PORTB=0xF9;
	   	TRISB = PORT_B_TECLA; 
	    TRISA = PORT_A_TECLA; 	
	}
	    
/*
 * Next: Was knob twisted?
 */

 	    if (Twisted) {
		apagado=0;
        TRISB = PORT_B_OUT;
        TRISA = PORT_A_OUT;
	 	twist_event ();
	    bcd_display ();
		save_parameters();
		PORTB=0xF9;
    	TRISB = PORT_B_TECLA; 	
        TRISA = PORT_A_TECLA; 
	    continue;
 	    }

/*
 * Next: Was keyboard pushed?
 */
 	    if (((PORTB&0xF0)^0xF0) || !PB4 || !PB5) {
		 	button_event ();
	        TRISB = PORT_B_OUT;
            TRISA = PORT_A_OUT;
		  	bcd_display ();
   			save_parameters();
			PORTB=0xF9;
    		TRISB = PORT_B_TECLA; 
            TRISA = PORT_A_TECLA; 
	        continue;
 	    }
/*
 * Next: If s-meter change update it
 */

	if (PB3) apagado=1;
    TRISB = PORT_B_OUT;
    TRISA = PORT_A_OUT;
    tmp=lee_adc();
	if (tmp>11) tmp=11;
	if (tmp!=ant_tmp) {
		adc_slope++;
		if (adc_slope>20) {
   		ant_tmp=tmp;
        PORTB = 0; 
	    lcd_write_instr (0xc2);
		    for (i=0;i<12;i++) {
				if (i>=tmp) lcd_write_data (' ');
				else lcd_write_data (0xFF);
			  }
		lcd_set_cursor (cur_pos[cursor_digit]);
		adc_slope=0;
		}
	}
	PORTB=0xF9;
	TRISB = PORT_B_TECLA; 
    TRISA = PORT_A_TECLA; 
        }
        
} /* end main */

//Main adjust program
void adjust_rutine(void) {
		cursor_digit=0;
        TRISB = PORT_B_OUT;
        TRISA = PORT_A_OUT;
        lcd_set_cursor (16);
        lcd_write_data ('A');
		load_bcd_packet(vfo,AD_FI+(mode*4));
		PORTB=0xF9;
    	TRISB = PORT_B_TECLA; 	
        TRISA = PORT_A_TECLA; 
		for (;;) {
			if (PB3) apagado=1;
	 	    if (Twisted) {
			 	twist_event ();
		        TRISB = PORT_B_OUT;
		        TRISA = PORT_A_OUT;
			  	bcd_display ();
				PORTB=0xF9;
		    	TRISB = PORT_B_TECLA; 	
		        TRISA = PORT_A_TECLA; 
		        continue;
	 	    }
	 	    if (((PORTB&0xF0)^0xF0) || !PB4 || !PB5) {
				delay(80);
				if (!PB5) {
			        TRISB = PORT_B_OUT;
			        TRISA = PORT_A_OUT;
					save_bcd_packet(vfo,AD_FI+(mode*4));
		     	    lcd_set_cursor (16);
			        lcd_write_data ('G');
		    		TRISB = PORT_B_TECLA; 
		 	        TRISA = PORT_A_TECLA; 
			        continue;
				}
				else {
					if (!PB1) {
						mode++;
						if (mode>4) mode=0;
						load_bcd_packet(vfo,AD_FI+(mode*4));
					}
					if (!PB3) apagar();
			        TRISB = PORT_B_OUT;
		            TRISA = PORT_A_OUT;
			        lcd_set_cursor (16);
			        lcd_write_data ('A');
				  	bcd_display ();
					PORTB=0xF9;
		    		TRISB = PORT_B_TECLA; 
	  	            TRISA = PORT_A_TECLA; 
			        continue;
				}
 	    	}
		}
	}		
// read internal adc by succesive aproximations
unsigned char lee_adc(void) {
unsigned char i,tmp;

    PORTA=0;
	TRISA|=0x0F;	
// comparador on
	CM2=0;
	CM1=1;
	CM0=0;
	delay(1);

// voltage reference on
	VREN=1;
    VROE=0;
	VRR=0;
	i=8;
	tmp=0;
//succesive aproximations algorithm
	while (i>0) {
		VRCON&=0xF0;
		VRCON|=tmp+i;
		if (!C1OUT) tmp+=i;
		i>>=1;
	}
// all off
  VREN=0;
  PORTA = 0;
  CM2=1;
  CM1=1;
  CM0=1;
  TRISA=PORT_A_OUT;
return tmp;
}


/*------------------------------------------------------------------
 *
 * Name:        twist_event
 *
 * Purpose:  	Handle the rotary encoder rotation events.   
 * 
 *------------------------------------------------------------------*/

void twist_event (void)
{
	
/*
 * Change frequency
 */
	    if (Twisted > 0) {		/* Clockwise. */
			bcd_incr (vfo, cursor_digit);
	        Twisted--;
	    }
	    else {
			bcd_decr (vfo, cursor_digit);
	        Twisted++;
	    }

}


/*------------------------------------------------------------------
 *
 * Name:        bcd_clear
 *
 * Purpose:     Set binary coded decimal string to zero.
 *
 * Inputs:      bcd	- Pointer to array of DIGITS bytes.
 *
 * Outputs:	Each element of bcd array is set to zero.
 *
 *------------------------------------------------------------------*/

void bcd_clear (byte_t *bcd)
{
	byte_t n;

	for (n=0; n<DIGITS; n++) {
	    bcd[n] = 0;
	}
}

/*------------------------------------------------------------------
 *
 * Name:        bcd_incr
 *
 * Purpose:     Increment specified value of BCD array.
 *
 * Inputs:      bcd	- Pointer to array of DIGITS bytes.
 *		position - Index of digit to increment.
 *			0 for the least significant byte.
 *
 * Outputs:	Digit(s) in BCD array are modified.
 * 
 * Description: If a digit overflows past 9, it is set to 0 and
 *		the next higher digit is incremented.
 *		Stop if number is >= 20,000,000.
 *
 * Example for DIGITS == 8:
 *	
 *		Initial contents of bcd: 	0 0 9 0 0 0 0 0
 *		which means 900 Hz.
 *		Position is 2 which means add 100 Hz.
 *
 *		Resulting contents of bcd:	0 0 0 1 0 0 0 0
 *
 * Example for DIGITS == 9:
 *	
 *		Initial contents of bcd: 	0 0 9 0 0 0 0 0 0
 *		which means 900.0 Hz.
 *		Position is 3 which means add 100 Hz.
 *
 *		Resulting contents of bcd:	0 0 0 1 0 0 0 0 0
 *
 *------------------------------------------------------------------*/


void bcd_incr (unsigned char bcd[], byte_t position)
{
	for ( ; position < DIGITS; position++) {
	    if (++bcd[position] <= 9) {
			break;
		}
		bcd[position]=0;
	} 

	if ((bcd[DIGITS-1] == 1) && (bcd[DIGITS-2] >= 5)) {
	    bcd_clear (bcd);		/* Don't go beyond 150 MHz. */
	    bcd[DIGITS-1] = 1; bcd[DIGITS-2] = 5;
	} 
}


/*------------------------------------------------------------------
 *
 * Name:        bcd_decr
 *
 * Purpose:     Decrement specified digit of BCD array.
 *
 * Inputs:      bcd	- Pointer to array of DIGITS bytes.
 *		position - Index of digit to decrement.
 *			0 for the least significant byte.
 *
 * Outputs:	Digit(s) in BCD array are modified.
 * 
 * Description: If a digit underflows below 0, it is set to 9 and
 *		the next higher digit is decremented.
 *		Stop if number reaches zero.
 *
 * Example for DIGITS == 8:	
 *
 *		Initial contents of bcd: 	0 0 0 1 0 0 0 0
 *		which means 1000 Hz.
 *		Position is 2 which means subtract 100.0 Hz.
 *
 *		Resulting contents of bcd:	0 0 9 0 0 0 0 0
 *
 * Example for DIGITS == 9:	
 *
 *		Initial contents of bcd: 	0 0 0 1 0 0 0 0 0
 *		which means 1000.0 Hz.
 *		Position is 3 which means subtract 100.0 Hz.
 *
 *		Resulting contents of bcd:	0 0 9 0 0 0 0 0 0
 *
 *------------------------------------------------------------------*/


void bcd_decr (unsigned char bcd[], byte_t position) {

	for ( ; position < DIGITS; position++) {
	    if (bcd[position] > 0) {
		bcd[position]--;
		return;
	    } 
		bcd[position]=9;
	}
	bcd_clear (bcd); 
	/* Oops, no more digits to "borrow from." */
	/* In other words we went past zero. */
}



/*------------------------------------------------------------------
 *
 * Name:        bcd_display
 *
 * Purpose:     Display number on the LCD.
 *
 * Inputs:      bcd	- Pointer to array of DIGITS bytes.
 *				cursor_digit - Index of digit to highlight with cursor.
 *				0 for the least significant byte.
 *
 * Outputs:	Digits are written to LCD.
 *		For readability, leading zeros are normally
 *		blanked and commas are inserted.
 * 
 * Example:	BCD array in:  		4 3 2 1 0 0 0 0
 *		cursor digit:	 	1
 *
 *		Displayed as:		1,234
 *					   ^
 *
 *------------------------------------------------------------------*/

void bcd_display (void)
{
/*
 * Position on LCD for each of the digits.
 * They are not adjacent because commas are inserted.
 */
	signed char d;
    byte_t leading_zero_blank;

	lcd_hide_cursor ();	
	leading_zero_blank = TRUE;
	lcd_set_cursor (0);
 	lcd_write_data (mode_t[mode][0]);
 	lcd_write_data (mode_t[mode][1]);
 	lcd_write_data (mode_t[mode][2]);
 	lcd_write_data (' ');

for (d = DIGITS-1; d >=0; d--) {
/*
 * Don't blank out digit at cursor.
 */
	    if (d == cursor_digit)
	        leading_zero_blank = FALSE;

/*
 * Don't blank out units (1 Hz) digit.
 */
        if (d == 0)
	        leading_zero_blank = FALSE;
/*
 * Non-zero digit found: stop blanking.
 */
	    if (vfo[d]) 
	        leading_zero_blank = FALSE;

		if (((d==3) || (d==0)))
			lcd_write_data('.');

	    if (leading_zero_blank) 
		 	lcd_write_data (' ');
	    else 
			lcd_write_data (vfo[d] + '0');
	    /* Blank any ".0" if cursor not set to last digit. */
	}
 	lcd_write_data (' ');
	// write VFO operation mode
 	lcd_write_data (oper_t[oper]);
	lcd_set_cursor (30);
	// write channel
 	lcd_write_data ((canal/10)%10+'0');
 	lcd_write_data (canal%10+'0');
	// put underline
	lcd_set_cursor (cur_pos[cursor_digit]);
	lcd_show_cursor ();	
	// update PLL
	add_fi(temp,vfo);
	pll_from_bcd(temp);	 
}

void add_fi (unsigned char bcd[], unsigned char valor[]){
unsigned char i,acarreo,tmp;

  acarreo=0;
  for (i=0;i<DIGITS;i++) {
	tmp=valor[i]+freqin[i]+acarreo;
	if (tmp>9) {
		acarreo=1;
		bcd[i]=tmp-10;		
		}
	else {
		acarreo=0;
		bcd[i]=tmp;		
	}
  }
}

/*------------------------------------------------------------------------
 *
 * Name:        delay
 *
 * Purpose:     Delay specified number of milliseconds.
 *		Not tested or guaranteed to be at all accurate.
 *
 *------------------------------------------------------------------------*/

void delay (byte_t ms)
{
        byte_t i;

        while (ms > 0) {

/*
 * Assuming 4 MHz clock, 1 instruction = 1 uS.
 * Loop has overhead of 3 instructions each time around.
 * One more instruction (NOP) should give total of 4 instructions
 * each time around the inner loop.
 */

            i = 249;
           
            do {
                NOP();
            } while (--i);

            ms--;
        }
}



/*------------------------------------------------------------------
 *
 * Name:        lcd_init
 *
 * Purpose:     Initialize the LCD.
 *
 * Description: There are a couple tricky things here.
 *              (1) A one line x 16 character display is wired
 *                  as two lines of 8 characters.
 *                  The second half begins at address 0x40.
 *              (2) We only have a 4 bit data path.  We start in
 *                  8 bit mode and switch to 4 bit mode.
 *
 *------------------------------------------------------------------*/

void lcd_init ()
{

/*
 * I don't understand the reason for the same command three
 * times with different delays but that's what the data
 * sheet says.
 */
        delay (60);             /* >= 15 mS. */
        lcd_put4 (3);
        delay (16);             /* >= 4.1 mS. */
        lcd_put4 (3);
        delay (1);              /* >= 0.1 mS. */
        lcd_put4 (3);
        delay (16);             /* >= 4.1 mS. */
        lcd_put4 (2);           /* Enter 4 bit data mode. */
        delay (1);

/*
 * We are now in 4 bit mode and must do writes in pairs.
 */
        lcd_put4 (2);           /* 4 bit mode again. */
        lcd_put4 (8);           /* 2 line display, 5x7 characters. */
        delay (1);

        lcd_put4 (2);           /* 4 bit mode again. */
        lcd_put4 (8);           /* 2 line display, 5x7 characters. */
        delay (1);

        lcd_put4 (2);           /* 4 bit mode again. */
        lcd_put4 (8);           /* 2 line display, 5x7 characters. */
        delay (1);


        lcd_put4 (0);           /* Display off. */
        lcd_put4 (8);
        delay (1);

        lcd_put4 (0);           /* Display on, Cursor and blinking off. */
        lcd_put4 (1);
        delay (3);              /* >= 1.64 mS. */

        lcd_put4 (0);           /* Cursor right automatically. */
        lcd_put4 (6);
        delay (3);

        lcd_put4 (0);           /* Display on, Cursor and blinking off. */
        lcd_put4 (12);
        delay (3);              /* >= 1.64 mS. */

}


/*------------------------------------------------------------------
 *
 * Name:        lcd_put4
 *
 * Purpose:     Send a nybble to the LCD with instruction mode.
 *              This is used only during initialization.
 *
 * Inputs:      x       - Only 4 low order bits used.
 *
 * Description: Write the data with,
 *                RS = 0 for instruction
 *                RW = 0 for write
 *              Pulse E.
 *
 * Note:        Initially we are in 8 bit mode and we only
 *              write the high order half of a full byte.
 *              After switching to 4 bit mode, this must be
 *              called in pairs, high order then low order half.
 *
 *------------------------------------------------------------------*/

void lcd_put4 (byte_t x)
{
        PORTB &= 0x0F;
        PORTB |= (x & 0x0F)<<4;
        NOP();
        LCD_e = 1;             
        NOP();
        LCD_e = 0;              
}


/*------------------------------------------------------------------
 *
 * Name:        lcd_write_data
 *
 * Purpose:     Send a character to display.
 *
 * Inputs:      x       - An ASCII character.
 *
 * Description: Write the data as two nybbles with
 *                RS set to 1 for data
 *                RW set to 0 for write
 *
 *------------------------------------------------------------------*/

void lcd_write_data (byte_t x)
{
	delay(1);
	LCD_rs = 1;
	PORTB &= 0x0F;
    PORTB |= x&0xF0;
    NOP();
    LCD_e = 1;             
    NOP();
    LCD_e = 0;              
	PORTB &= 0x0F;
    PORTB |= x<<4;
    NOP();
    LCD_e = 1;             
    NOP();
    LCD_e = 0;
}



/*------------------------------------------------------------------
 *
 * Name:        lcd_write_instr
 *
 * Purpose:     Send an instruction to display.
 *
 * Inputs:      x       - Instruction code.
 *
 * Description: Write the data as two nybbles with
 *                RS set to 0 for instruction
 *                RW set to 0 for write
 *
 * 		The application shouldn't use this function.
 *		Instead it should use functions such as 
 *		lcd_set_cursor.
 *
 *------------------------------------------------------------------*/

void lcd_write_instr (byte_t x)
{
    LCD_e = 0;              
	delay(2);
	LCD_rs = 0;
	PORTB &= 0x0F;
        PORTB |= x&0xF0;
        NOP();
        LCD_e = 1;             
        NOP();
        LCD_e = 0;              
	PORTB &= 0x0F;
        PORTB |= x<<4;
        NOP();
        LCD_e = 1;             
        NOP();
        LCD_e = 0;
}



/*------------------------------------------------------------------
 *
 * Name:        lcd_set_cursor
 *
 * Purpose:     Set memory address in preparation for a data write
 *		or displaying the cursor.
 *
 * Inputs:      position	- 0 for left .. 15 for right edge.
 *
 * Description: Note that a 16 character display has memory organized
 *		as two non-adjacent 8 character areas.
 *		This function hides that detail so the rest of the
 *		application doesn't have to be concerned about it.
 *
 *		The commands to set the address are:
 *
 *              1 0 0 0 0 n n n         for first 8 characters
 *              1 1 0 0 0 n n n         for second 8 characters.
 *
 *------------------------------------------------------------------*/

void lcd_set_cursor (byte_t position)  
// Note that: two lines, first from 0 to 15, second from 16 to 31
{
	if (position < 16) {
	    lcd_write_instr (0x80 | position);
	}
	else {
	    lcd_write_instr (0xc0 | (position-16));
	}
}


/*------------------------------------------------------------------
 *
 * Name:        lcd_hide_cursor
 *		lcd_show_cursor
 *
 * Purpose:     Turn cursor display off / on.   
 *
 *------------------------------------------------------------------*/

void lcd_hide_cursor (void)
{
	lcd_write_instr (0x0C);
}

#if 1			/* Using underscore cursor. */
			/* Requires at least 8 vertical dots */
			/* per character.  Some have only 7. */
 
void lcd_show_cursor (void)
{
	lcd_write_instr (0x0E);
}

#else			/* Using blinking. */
 
void lcd_show_cursor (void)
{
	lcd_write_instr (0x0D);
}

#endif


/*------------------------------------------------------------------
 *
 * Name:        pll_init
 *
 * Purpose:     Initialize the PLL chip.   
 *
 *
 *------------------------------------------------------------------*/

void pll_init (void)
{
unsigned char i,tmp;
unsigned int tmpl;
// PLL setup routine, only needs to be done once, at startup.


	//----- Initilize PLL -----
	// EN High, Din Low, 4 clocks
	CLK = 0;
	DIN = 0;
	EN = 1;

	for (i=0;i<4;i++) {

		CLK = 1;
		CLK = 0;	
	}
	// EN Low, Din Low (except clock 4 = 1), 5 clocks
	DIN = 0;
	EN = 0;

	for (i=0;i<5;i++) {

		CLK = 1;
		CLK = 0;	
	}

	EN = 1;
	// PLL is now reset
	
	//----- Setup C Reg with proper bits -----
	EN = 0;
	tmp=0x80; 
	// value to program: 0000 0001
	for( i = 0 ; i < 8; i++)
	{
		// Decompose the PLL N value into bits, set data pin
	    if (0x01&tmp) DIN=1;
	    else DIN=0;
		tmp>>=1;
		// Cycle the clock
		CLK = 1;
		CLK = 0;	
	}
	
	// EN back high
	EN = 1;

	//----- Setup Ref Divider -----
	// 10 MHz / 3200 = 3.2 KHz (3200 = 0x0c80)
	EN = 0;

	tmpl=0x4000;
	for( i = 0 ; i < 15; i++)
	{
		// Decompose the PLL N value into bits, set data pin
	    if (0x0c80U&tmpl) DIN=1;
	    else DIN=0;
		tmpl>>=1;
		// Cycle the clock
		CLK = 1;
		CLK = 0;	
	}

	// Enable Back High
	EN = 1;
}

/*------------------------------------------------------------------
 *
 * Name:        pll_from_bcd
 *
 * Purpose:     Convert frequency, in BCD, to binary then send
 *		it to the PLL chip.  
 *
 * Inputs:      bcd	- Pointer to array of 'DIGITS' digits.     
 *
 * Description: Convert from BCD to binary by multiplying each
 *		digit by an appropriate constant and adding them
 *		all up. 
 *
 *------------------------------------------------------------------*/

void pll_from_bcd (unsigned char *bcd)
{
unsigned int a,band;
unsigned long tmp;
unsigned char i,b;

 if (bcd[6]) tmp=1000000;
 else tmp=0;
 for (i=0;i<bcd[5];i++)
	tmp+=100000;
 for (i=0;i<bcd[4];i++)
	tmp+=10000;
 for (i=0;i<bcd[3];i++)
	tmp+=1000;
 for (i=0;i<bcd[2];i++)
	tmp+=100;
 for (i=0;i<bcd[1];i++)
	tmp+=10;
 tmp+=bcd[0];

 b=((unsigned char)tmp)&0x1F;
 a=(unsigned int)((tmp)>>5); //division by 32
 // output coarse tuning
 pll_write_freq(a);
 // output fine tuning
 dac_write(b);
 // band calculation
 if (vfo[6]) banda=6;
 else if (vfo[5]>4) banda=5;
 else if (vfo[5]>1) {
	if (vfo[4]>4) banda=4;
    else banda=3;
  }
 else if (vfo[5]>0) banda=2;
 else if (vfo[4]>6) banda=1;
 else banda=0;

 band=1<<banda;
 band<<=8;
 band|=1<<mode; 
 //output band and mode
 bpf_write(band); 
}

/*------------------------------------------------------------------
 *
 * Name:        pll_write_freq
 *
 * Purpose:     Send frequency control information to DDS chip.  
 *
 * Inputs:      f       - Pointer to LSB of 4 data bytes.     
 *
 * Description: Shift out 4 bytes of frequency and 1 of phase/control.
 *
 *------------------------------------------------------------------*/

 
void pll_write_freq (unsigned int f)
{
unsigned char ctr;
unsigned int tmp;

// Set clock, enable low
    tmp=0x8000U;	
	// Shift in channel frequency to N reg (MSB first)
	CLK = 0;
    DIN = 0;
	EN = 0;
	for( ctr = 0 ; ctr < 16 ; ctr++)
	{
		// Decompose the PLL N value into bits, set data pin
	    if (f&tmp) DIN=1;
	    else DIN=0;
		tmp>>=1;
		// Cycle the clock
		CLK = 1;
		CLK = 0;	
	}
	
	// Clean up and set enable high which 
	// transfers count to N reg
	DIN = 0;
	EN = 1;
}

/*------------------------------------------------------------------
 *
 * Name:        dac_write
 *
 * Purpose:     Set an output of the quad D/A converter.
 *
 * Inputs:      addr    - Desired output.  One of:
 *
 *              value   - Data for the DAC, 0 - 255.
 *
 * Description: 
 *
 *------------------------------------------------------------------*/

void dac_write (byte_t f)
{
 char ctr;
unsigned char tmp;
 f=(f<<3)+(apaga<<1);
// Set clock, enable low
	CLK = 0;
	EN_DAC = 0;
    tmp=0x80;	
	// Shift in (MSB first)
	for( ctr = 0 ; ctr < 8 ; ctr++)
	{
		// Decompose the value into bits, set data pin
	    if (f&tmp) DIN=1;
	    else DIN=0;
		tmp>>=1;
		// Cycle the clock
		CLK = 1;
		CLK = 0;	
	}
	
	// Clean up and set enable high which 
	EN_DAC = 1; 
	delay(1);
	EN_DAC = 0; 
	DIN = 0;
}

 void bpf_write (unsigned int f)
{
 char ctr;
unsigned int tmp;
// Set clock, enable low
	CLK = 0;
	EN_BPF = 0;
    tmp=0x8000U;	
	// Shift in (MSB first)
	for( ctr = 0 ; ctr < 16 ; ctr++)
	{
		// Decompose the value into bits, set data pin
	    if (f&tmp) DIN=1;
	    else DIN=0;
		tmp>>=1;
		// Cycle the clock
		CLK = 1;
		CLK = 0;	
	}
	
	// Clean up and set enable high which 
	// transfers count to N reg
	DIN = 0;
	EN_BPF = 1; 
	delay(1);
	EN_BPF = 0; 
} 
	    
/*------------------------------------------------------------------
 *
 * Name:        enc_sample
 *
 * Purpose:     Sample the rotary encoder and look for rotation.
 *
 * Inputs:	Two bits from an input port.
 *
 * Gobal Out:	Twisted		- which is used by main loop.
 *				  Increment for right rotation.
 *				  Decrement for left rotation.
 *
 *------------------------------------------------------------------*/

#define MIN_STABLE 12

byte_t prev_sample;
byte_t stable_count;
byte_t prev_debounced;
 
void enc_sample (void)
{
	byte_t sample,tmp;		
	byte_t debounced;
	byte_t rotation;
	

	sample = PORTB & 0x09;
	tmp=sample>>2;
	sample&=1;
	sample|=tmp;

/*
 * Debounce.
 * Don't use the value until it has been stable for
 * at least 1/100 second.  (24 samples in this case.)
 */
 
	if (sample == prev_sample) {
	    if (stable_count < MIN_STABLE) {
		stable_count++;
		if (stable_count == MIN_STABLE) {
/*
 * Convert from Gray code to binary.
 */
		    debounced = sample ^ (sample >> 1);
		}
	    }
	    else {
		return;
	    }
	} 
	else {
	    stable_count = 0;
	    prev_sample = sample;
	    return;
	}

/*
 * Was there any change?
 * Careful!  In some cases you get the expected +1 or -1
 * but only the two low order bits are valid here.
 */
 	rotation = debounced - prev_debounced;

 	if (rotation) {
	    if (rotation & 2) {
	        if (Twisted<4) Twisted++;
	    }
	    else {
		if (Twisted>-4) Twisted--;
	    }
	}

	prev_debounced = debounced;
}


/*------------------------------------------------------------------------
 *
 * Name:        data_data_eeprom_write
 *
 * Purpose:     Write a byte to non-volatile data memory.
 *
 * Inputs:      adr     - Address in range of 0 to 63.
 *              dat     - Data byte in range of 0 to 255.
 *
 * Outputs:     Write to non-volatile memory.
 *
 * Returns:     None.
 *
 * Assumptions: Address is within range.
 *              This program does not use interrupts.
 *
 * Description: This is a lot more complicated than read to
 *              avoid accidental writes.
 *              Set the EEADR and EEDATA registers.
 *              Write certain patterns to EECON2.
 *              Wait for completion which is typically 10 mS.
 *
 *------------------------------------------------------------------------*/

void data_eeprom_write (byte_t adr, byte_t dat)
{
        EEADR = adr;
        EEDATA = dat;
        GIE = 0; /* if using interrupts. */
        WREN = 1;
        EECON2 = 0x55;
        EECON2 = 0xaa;
        WR = 1;
        WREN = 0;
        GIE = 1; /* if using interrupts. */
        while (WR)
            ;
        EEIF = 0;
}

// Load frequency and mode in EEPROM pos
void load_bcd_packet(unsigned char bcd[],unsigned char pos) {
unsigned char i,tmp;
for (i=0;i<3;i++) {
	tmp=EEPROM_READ(pos+i);
	bcd[6-(i<<1)]=tmp>>4;
	bcd[5-(i<<1)]=tmp&0x0F;
	}
    bcd[0]=EEPROM_READ(pos+3)&0x0F;
// Two formats, 4 or 5 bytes 
if (mod_packet) {
	tmp=EEPROM_READ(pos+4);
	mode=tmp&0x7;
  }
}

// Store frequency and mode in EEPROM pos
void save_bcd_packet(unsigned char bcd[],unsigned char pos) {
unsigned char i,tmp;

for (i=0;i<3;i++) {
	tmp=bcd[6-(i<<1)]<<4;
	tmp|=bcd[5-(i<<1)]&0x0F;
	data_eeprom_write(pos+i,tmp);
	}
	data_eeprom_write(pos+3,bcd[0]);
// Two formats, 4 or 5 bytes 
if (mod_packet) 
	data_eeprom_write(pos+4,mode); //+(banda<<3));
}

// Load all parameters at power up
void load_parameters(void){
signed char pos;

/* load vfos and call frequency */
canal=EEPROM_READ(AD_CANAL);
oper=EEPROM_READ(AD_OPER);
cursor_digit=EEPROM_READ(AD_CURSOR);
switch (oper) {
	case OPER_VFOA: pos=AD_VFOA;break;
	case OPER_VFOB: pos=AD_VFOB;break;
	case OPER_CANAL: pos=AD_CHNL+canal*5;break;
	case OPER_CALL: pos=AD_CALL;break;
	case OPER_SPLIT: pos=AD_VFOA;break;
}
mod_packet=1;
load_bcd_packet(vfo,pos);
// Load FI programated value
mod_packet=0;
load_bcd_packet(freqin,AD_FI+(mode*4));
mod_packet=1;
}

// Initialize parameters stored in EEPROM
void init_parameters(void){
unsigned char i;

mode=2;
banda=2;
vfo[6]=0;
vfo[5]=1;
vfo[4]=0;
vfo[3]=0;
vfo[2]=0;
vfo[1]=0;
vfo[0]=0;
//Initialize IF values
mod_packet=0;
for (i=0;i<5;i++)
	save_bcd_packet(vfo,AD_FI+(i*4));
mod_packet=1;
//Initialize VFOs values and chanels
vfo[5]=1;
vfo[4]=4;
vfo[3]=1;
vfo[2]=0;
vfo[1]=0;
for (i=0;i<3;i++)
	save_bcd_packet(vfo,AD_VFOA+i*5);
for (i=0;i<17;i++)
	save_bcd_packet(vfo,AD_CHNL+i*5);
canal=0;
oper=0;
cursor_digit=0;
/* write the mark */
data_eeprom_write(AD_MARCA,MARCA);
save_parameters();
}

// Save all operatings parameters
void save_parameters(void) {
signed char pos;

/* write active VFO or calling frequency */
pos=-1;
switch (oper) {
	case OPER_VFOA: pos=AD_VFOA;break;
	case OPER_VFOB: pos=AD_VFOB;break;
	case OPER_CANAL: pos=AD_CHNL+(canal*5);break;
	case OPER_CALL: pos=AD_CALL;break;
	case OPER_SPLIT: break;
}
if (pos>=0) save_bcd_packet(vfo,pos);
// Store actual canal
data_eeprom_write(AD_CANAL,canal);
// Store actual VFO mode
data_eeprom_write(AD_OPER,oper);
// Store actual STEP 
data_eeprom_write(AD_CURSOR,cursor_digit);
}

// Interrupt Service Routine to process encoder rotation
 void interrupt timer (void) {
#if XTAL == 4000000
	TMR0 = 256-198; /* 256-198  ->  2404.2 Hz. */
					/* 256-197  ->  2415.9 Hz. */
 #else
#error Unexpected XTAL.
#endif
	T0IF = 0;
    enc_sample ();
} 

// Process button press
void button_event(void) {
unsigned char pos;
	// debounce button
	delay(80);
	// Process
 	if (!PB0) {
		// STEP change
		cursor_digit++;
		if (cursor_digit>5) cursor_digit=0;
		}
	else if (!PB1) {
		// MODE change
		mode++;
		if (mode>4) mode=0;
		// Reload FI value
		mod_packet=0;
		load_bcd_packet(freqin,AD_FI+(mode*4));
		mod_packet=1;
		}
	else if (!PB2) {
		// VFO operating mode change
		oper++;
		if (oper>4) oper=0;
		switch (oper) {
			case OPER_VFOA: pos=AD_VFOA;break;
			case OPER_VFOB: pos=AD_VFOB;break;
			case OPER_CANAL: pos=AD_CHNL+canal*5; break;
			case OPER_CALL: pos=AD_CALL;break;
			case OPER_SPLIT: pos=AD_VFOA;break;
		}
		// Reload VFO value
		load_bcd_packet(vfo,pos);
		}
	else if (!PB3) {
		// Power OFF!!
		apagar();
		}
	else if ((!PB4) || (!PB5)) {
		// Chanel up/down or band up/down
	  if (oper==OPER_CANAL) {
		if (!PB4) canal++;
		else canal--;
		if (canal<0) canal=16;
		if (canal>16) canal=0;
		//Reload chanell value
		load_bcd_packet(vfo,AD_CHNL+canal*5);
		}
	 else {
		if (!PB4) banda++;
		else banda--;
  		if (banda>5) banda=0;
		if (banda<0) banda=6;
		//Adjust band frequency
		vfo[6]=0;
		if (banda<2) {
				vfo[5]=0;				
				vfo[4]=lista_banda[banda]>>4;
				vfo[3]=lista_banda[banda]&0x0F;
				}
		else if (banda<6) {
				vfo[5]=lista_banda[banda]>>4;
				vfo[4]=lista_banda[banda]&0x0F;
		}
		else {
				vfo[6]=1;
				vfo[5]=4;
				vfo[4]=4;
		}
	}
  }
}

// To power off!!
void apagar(void) {
		if (apagado) {
        TRISB = PORT_B_OUT;
        TRISA = PORT_A_OUT;
		PORTB=0;
			// Power off display
		    lcd_write_instr (0x08);
			// Power off all
			apaga=0;
			dac_write(0);
			// Stay forever
			for(;;);
			}
}


/* END */
